import time
import unittest

from galileodb.factory import create_influxdb_from_env
from galileodb.model import ExperimentEvent, RequestTrace, Telemetry


class TestInfluxExperimentDatabase(unittest.TestCase):

    def setUp(self) -> None:
        self.exp_db = create_influxdb_from_env()
        self.exp_db.open()
        self.exp_id = 'exp-1'
        self.exp_bucket = self.exp_db.client.buckets_api().create_bucket(bucket_name=self.exp_id,
                                                                         org_id=self.exp_db.org_id)

    def tearDown(self) -> None:
        self.exp_db = create_influxdb_from_env()
        self.exp_db.open()
        self.exp_db.client.buckets_api().delete_bucket(self.exp_bucket.id)

    @unittest.skip("needs environment variables set - test manually")
    def test_save_event(self):
        exp_id = self.exp_id

        exp_db = self.exp_db

        # print(exp_db.get_events(exp_id))

        events = [
            ExperimentEvent(exp_id, time.time(), 'some-event', 'some-value'),
            ExperimentEvent(exp_id, time.time(), 'some-event', 'some-value'),
            ExperimentEvent(exp_id, time.time(), 'some-event', 'some-value'),
            ExperimentEvent(exp_id, time.time(), 'some-event', 'some-value'),
            ExperimentEvent(exp_id, time.time(), 'some-event', 'some-value'),
        ]
        exp_db.save_events(events)

        for event in events:
            exp_db.save_events([event])

        retrieved_events = exp_db.get_events(exp_id)
        self.assertEquals(events, retrieved_events)

    @unittest.skip("needs environment variables set - test manually")
    def test_save_traces(self):
        exp_id = self.exp_id
        exp_db = self.exp_db

        original_traces = [
            RequestTrace('1', 'client-1', 'resnet', time.time(), time.time() + 0.01, time.time() + 0.1, 200, 'server-1',
                         exp_id, '{"headers": 1}', 'response'),
            RequestTrace('1', 'client-1', 'resnet', time.time(), time.time() + 0.01, time.time() + 0.1, 200, 'server-1',
                         exp_id, response='response'),
            RequestTrace('1', 'client-1', 'resnet', time.time(), time.time() + 0.01, time.time() + 0.1, 200, 'server-1',
                         exp_id, response='response'),
            RequestTrace('1', 'client-1', 'resnet', time.time(), time.time() + 0.01, time.time() + 0.1, 200, 'server-1',
                         exp_id, response='response'),
            RequestTrace('1', 'client-1', 'resnet', time.time(), time.time() + 0.01, time.time() + 0.1, 200, 'server-1',
                         exp_id, response='response')
        ]

        for trace in original_traces:
            exp_db.save_traces([trace])

        traces = exp_db.get_traces(exp_id)
        print(traces)
        self.assertEquals(original_traces, traces, 'traces are not equal')

    @unittest.skip("needs environment variables set - test manually")
    def test_save_telemetry(self):
        exp_id = self.exp_id
        exp_db = self.exp_db

        original_telemetry = [
            Telemetry(timestamp=time.time(), metric='cpu', node='node-0', value=0, exp_id=exp_id, subsystem="core1"),
            Telemetry(timestamp=time.time(), metric='cpu', node='node-0', value=1, exp_id=exp_id),
            Telemetry(timestamp=time.time(), metric='cpu', node='node-0', value=2, exp_id=exp_id),
            Telemetry(timestamp=time.time(), metric='cpu', node='node-0', value=3, exp_id=exp_id),
            Telemetry(timestamp=time.time(), metric='cpu', node='node-0', value=-1, exp_id=exp_id)
        ]

        for tel in original_telemetry:
            exp_db.save_telemetry([tel])

        telemetry = exp_db.get_telemetry(exp_id)

        self.assertEquals(original_telemetry, telemetry, 'telemetry are not equal')



from unittest import TestCase

from timeout_decorator import timeout_decorator

from galileodb.model import RequestTrace
from galileodb.reporter.traces import RedisTraceReporter
from tests.testutils import RedisResource, RedisSubscriber


class TestRedisTraceReporter(TestCase):
    redis = RedisResource()

    def setUp(self) -> None:
        self.redis.setUp()

    def tearDown(self) -> None:
        self.redis.tearDown()

    @timeout_decorator.timeout(5)
    def test_report_multiple(self):
        reporter = RedisTraceReporter(self.redis.rds)
        subscriber = RedisSubscriber(self.redis.rds, RedisTraceReporter.channel)
        subscriber.start()

        fixture = [
            RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, response='hello there'),
            RequestTrace('r2', 'c1', 's1', 2.1, 2.2, 2.3, 200, server='server'),
        ]

        reporter.report_multiple(fixture)

        try:
            data = subscriber.queue.get(timeout=1)
            self.assertTrue('r1' in data)
            self.assertTrue('r2' not in data)
            self.assertTrue('server' not in data)
            self.assertTrue('hello there' in data)

            data = subscriber.queue.get(timeout=1)
            self.assertTrue('r1' not in data)
            self.assertTrue('r2' in data)
            self.assertTrue('server' in data)
            self.assertTrue('hello there' not in data)

            self.assertTrue(subscriber.queue.empty())
        finally:
            subscriber.shutdown()

    @timeout_decorator.timeout(5)
    def test_report_response_with_noisy_string(self):
        reporter = RedisTraceReporter(self.redis.rds)
        subscriber = RedisSubscriber(self.redis.rds, RedisTraceReporter.channel)

        subscriber.start()

        response = "foo=bar\nfield1=value1,a,b,c"

        reporter.report_multiple([
            RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, response=response),
        ])

        data = subscriber.queue.get(1)
        self.assertEqual(r"r1,c1,s1,1.1000000,1.2000000,1.3000000,200,None,None,None,foo=bar\nfield1=value1,a,b,c",
                         data)

        subscriber.shutdown()

    @timeout_decorator.timeout(5)
    def test_report_with_heaers(self):
        reporter = RedisTraceReporter(self.redis.rds)
        subscriber = RedisSubscriber(self.redis.rds, RedisTraceReporter.channel)

        subscriber.start()

        headers = '{"server": ["Server", "BaseHTTP/0.6 Python/3.9.5"], "date": ["Date", "Tue, 14 Sep 2021 08:54:40 GMT"], "content-type": ["Content-type", "text/html"]}'
        expected = RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, response='foo=bar\nfield1=value1,a,b,c',
                                headers=headers)

        reporter.report_multiple([
            expected,
        ])

        data = subscriber.queue.get(1)
        self.assertEqual(
            fr"r1,c1,s1,1.1000000,1.2000000,1.3000000,200,None,None,{headers.replace(',', '|')},foo=bar\nfield1=value1,a,b,c",
            data)

        subscriber.shutdown()



import time
import unittest

from galileodb.model import Event, ExperimentEvent
from galileodb.recorder.events import ExperimentEventRecorder, ExperimentEventRecorderThread, \
    BatchingExperimentEventRecorder
from galileodb.reporter.events import RedisEventReporter
from galileodb.sql.adapter import ExperimentSQLDatabase
from tests.testutils import RedisResource, SqliteResource


class TestExperimentEventRecorder(unittest.TestCase):
    redis_resource: RedisResource = RedisResource()
    db_resource: SqliteResource = SqliteResource()

    exp_db: ExperimentSQLDatabase
    reporter: RedisEventReporter

    def setUp(self) -> None:
        self.redis_resource.setUp()
        self.db_resource.setUp()
        self.reporter = RedisEventReporter(self.redis_resource.rds)

    def tearDown(self) -> None:
        self.redis_resource.tearDown()
        self.db_resource.tearDown()

    def test_batching_recorder_records_correctly(self):
        thread = ExperimentEventRecorderThread(
            BatchingExperimentEventRecorder(self.redis_resource.rds, self.db_resource.db, 'unittest', flush_every=1)
        )
        thread.start()

        time.sleep(0.1)

        thread.recorder._record(Event(1., 'start', 'function1'))
        thread.recorder._record(Event(2., 'stop', 'function1'))
        thread.recorder._record(Event(3., 'exit'))
        thread.stop()

        records = self.db_resource.sql.fetchall('SELECT * FROM `events` WHERE EXP_ID = "unittest"')
        self.assertEqual(3, len(records))

        self.assertEqual(ExperimentEvent('unittest', 1., 'start', 'function1'), records[0])
        self.assertEqual(ExperimentEvent('unittest', 2., 'stop', 'function1'), records[1])
        self.assertEqual(ExperimentEvent('unittest', 3., 'exit'), records[2])

    def test_batching_recorder_flush_after_stop(self):
        thread = ExperimentEventRecorderThread(
            BatchingExperimentEventRecorder(self.redis_resource.rds, self.db_resource.db, 'unittest', flush_every=5)
        )
        thread.start()

        time.sleep(0.1)

        thread.recorder._record(Event(1., 'start', 'function1'))
        thread.recorder._record(Event(2., 'stop', 'function1'))
        thread.recorder._record(Event(3., 'exit'))

        thread.stop()

        records = self.db_resource.sql.fetchall('SELECT * FROM `events` WHERE EXP_ID = "unittest"')
        self.assertEqual(3, len(records))

        self.assertEqual(ExperimentEvent('unittest', 1., 'start', 'function1'), records[0])
        self.assertEqual(ExperimentEvent('unittest', 2., 'stop', 'function1'), records[1])
        self.assertEqual(ExperimentEvent('unittest', 3., 'exit'), records[2])

    def test_batching_recorder_with_redis(self):
        thread = ExperimentEventRecorderThread(
            BatchingExperimentEventRecorder(self.redis_resource.rds, self.db_resource.db, 'unittest', flush_every=1)
        )
        thread.start()

        time.sleep(0.5)

        self.redis_resource.rds.publish("galileo/events", "1. start function1")
        self.redis_resource.rds.publish("galileo/events", "2. stop function1")
        self.redis_resource.rds.publish("galileo/events", "3. exit")

        time.sleep(0.5)

        thread.stop()

        records = self.db_resource.sql.fetchall('SELECT * FROM `events` WHERE EXP_ID = "unittest"')
        self.assertEqual(3, len(records))

        self.assertEqual(ExperimentEvent('unittest', 1., 'start', 'function1'), records[0])
        self.assertEqual(ExperimentEvent('unittest', 2., 'stop', 'function1'), records[1])
        self.assertEqual(ExperimentEvent('unittest', 3., 'exit'), records[2])

    def test_recorder_with_redis(self):
        thread = ExperimentEventRecorderThread(
            ExperimentEventRecorder(self.redis_resource.rds, self.db_resource.db, 'unittest')
        )
        thread.start()

        time.sleep(0.5)

        self.redis_resource.rds.publish("galileo/events", "1. start function1")
        self.redis_resource.rds.publish("galileo/events", "2. stop function1")
        self.redis_resource.rds.publish("galileo/events", "3. exit")

        time.sleep(0.5)

        thread.stop()

        records = self.db_resource.sql.fetchall('SELECT * FROM `events` WHERE EXP_ID = "unittest"')
        self.assertEqual(3, len(records))

        self.assertEqual(ExperimentEvent('unittest', 1., 'start', 'function1'), records[0])
        self.assertEqual(ExperimentEvent('unittest', 2., 'stop', 'function1'), records[1])
        self.assertEqual(ExperimentEvent('unittest', 3., 'exit'), records[2])



from galileodb.model import Event


class RedisEventReporter:

    def __init__(self, rds) -> None:
        self.rds = rds

    def report(self, event: Event):
        channel = f'galileo/events'

        if event.value is None:
            msg = f'{event.timestamp} {event.name}'
        else:
            msg = f'{event.timestamp} {event.name} {event.value}'

        self.rds.publish(channel, msg)



from galileodb.recorder.events import ExperimentEventRecorderThread, ExperimentEventRecorder
from galileodb.recorder.telemetry import ExperimentTelemetryRecorder
from galileodb.recorder.traces import RedisTraceRecorder
from galileodb.trace import DatabaseTraceWriter


class Recorder:

    def __init__(self, rds, exp_db, experiment_id) -> None:
        self.rds = rds
        self.exp_db = exp_db
        self.experiment_id = experiment_id

        self.telemetry_recorder = ExperimentTelemetryRecorder(rds, exp_db, experiment_id)
        self.event_recorder = ExperimentEventRecorderThread(ExperimentEventRecorder(rds, exp_db, experiment_id))
        self.trace_recorder = RedisTraceRecorder(rds, experiment_id, DatabaseTraceWriter(exp_db))

    def start(self):
        self.telemetry_recorder.start()
        self.event_recorder.start()
        self.trace_recorder.start()

    def join(self, timeout=None):
        self.telemetry_recorder.join(timeout)
        self.event_recorder.join(timeout)
        self.trace_recorder.join(timeout)

    def stop(self, timeout=None):
        self.telemetry_recorder.stop(timeout)
        self.event_recorder.stop(timeout)
        self.trace_recorder.stop(timeout)



from typing import List

from galileodb import ExperimentDatabase, Experiment, NodeInfo, Telemetry
from galileodb.influx.db import InfluxExperimentDatabase
from galileodb.model import ExperimentEvent, RequestTrace
from galileodb.sql.adapter import ExperimentSQLDatabase


class MixedExperimentDatabase(ExperimentDatabase):
    """
    Implements the ExperimentDatabase using InfluxDB for telemetry, traces and events and
    SQL for experiment metadata
    """
    influxdb: InfluxExperimentDatabase
    sqldb: ExperimentSQLDatabase

    def __init__(self, influxdb: InfluxExperimentDatabase, sqldb: ExperimentSQLDatabase):
        self.influxdb = influxdb
        self.sqldb = sqldb

    def open(self):
        self.influxdb.open()
        self.sqldb.open()

    def close(self):
        self.influxdb.close()
        self.sqldb.close()

    def save_experiment(self, experiment: Experiment):
        self.sqldb.save_experiment(experiment)
        self.influxdb.client.buckets_api().create_bucket(
            bucket_name=experiment.id,
            org_id=self.influxdb.org_id
        )

    def update_experiment(self, experiment: Experiment):
        self.sqldb.update_experiment(experiment)

    def delete_experiment(self, exp_id: str):
        self.sqldb.delete_experiment(exp_id)

    def get_experiment(self, exp_id: str) -> Experiment:
        return self.sqldb.get_experiment(exp_id)

    def save_traces(self, traces: List[RequestTrace]):
        self.influxdb.save_traces(traces)

    def touch_traces(self, experiment: Experiment):
        self.sqldb.touch_traces(experiment)

    def get_traces(self, exp_id: str) -> List[RequestTrace]:
        return self.influxdb.get_traces(exp_id)

    def save_telemetry(self, telemetry: List[Telemetry]):
        self.influxdb.save_telemetry(telemetry)

    def get_telemetry(self, exp_id=None) -> List[Telemetry]:
        return self.influxdb.get_telemetry(exp_id)

    def save_event(self, event: ExperimentEvent):
        self.influxdb.save_event(event)

    def save_events(self, events: List[ExperimentEvent]):
        self.influxdb.save_events(events)

    def get_events(self, exp_id) -> List[ExperimentEvent]:
        return self.influxdb.get_events(exp_id)

    def save_nodeinfos(self, infos: List[NodeInfo]):
        self.sqldb.save_nodeinfos(infos)

    def find_all(self) -> List[Experiment]:
        return self.sqldb.find_all()

    def get_running_experiment(self) -> Experiment:
        return self.sqldb.get_running_experiment()



import logging
import time
import unittest

from telemc import Telemetry
from timeout_decorator import timeout_decorator

from galileodb.reporter.telemetry import RedisTelemetryReporter
from galileodb.sql.adapter import ExperimentSQLDatabase
from galileodb.recorder.telemetry import ExperimentTelemetryRecorder
from tests.testutils import RedisResource, SqliteResource, poll

logging.basicConfig(level=logging.DEBUG)


class TestExperimentTelemetryRecorder(unittest.TestCase):
    redis_resource: RedisResource = RedisResource()
    db_resource: SqliteResource = SqliteResource()

    exp_db: ExperimentSQLDatabase
    reporter: RedisTelemetryReporter

    def setUp(self) -> None:
        self.redis_resource.setUp()
        self.db_resource.setUp()
        self.reporter = RedisTelemetryReporter(self.redis_resource.rds)

    def tearDown(self) -> None:
        self.redis_resource.tearDown()
        self.db_resource.tearDown()

    @timeout_decorator.timeout(5)
    def test_recorder_records_correctly(self):
        recorder = ExperimentTelemetryRecorder(self.redis_resource.rds, self.db_resource.db, 'unittest', flush_every=1)

        recorder._record(Telemetry('1.0', '31', 'node1', 'cpu'))
        recorder._record(Telemetry('2.0', '32', 'node2', 'cpu'))
        recorder._record(Telemetry('3.0', '33', 'node2', 'rx', 'eth0'))

        records = self.db_resource.sql.fetchall('SELECT * FROM `telemetry` WHERE EXP_ID = "unittest"')
        self.assertEqual(3, len(records))
        self.assertEqual(('unittest', 1.0, 'cpu', None, 'node1', 31.0), records[0])
        self.assertEqual(('unittest', 2.0, 'cpu', None, 'node2', 32.0), records[1])
        self.assertEqual(('unittest', 3.0, 'rx', 'eth0', 'node2', 33.0), records[2])

    @timeout_decorator.timeout(5)
    def test_publish_non_float_value_does_not_break_recorder(self):
        recorder = ExperimentTelemetryRecorder(self.redis_resource.rds, self.db_resource.db, 'unittest')
        recorder.start()
        time.sleep(0.5)

        try:
            self.reporter.report(Telemetry('5', '35', 'node1', 'cpu'))
            self.reporter.report(Telemetry('6', 'foo', 'node1', 'cpu'))
            self.reporter.report(Telemetry('7', '37', 'node2', 'cpu'))
        finally:
            time.sleep(0.5)
            recorder.stop(timeout=2)

        records = self.db_resource.sql.fetchall('SELECT * FROM `telemetry` WHERE EXP_ID = "unittest"')
        self.assertEqual(2, len(records))
        self.assertEqual(('unittest', 5.0, 'cpu', None, 'node1', 35.0), records[0])
        self.assertEqual(('unittest', 7.0, 'cpu', None, 'node2', 37.0), records[1])


if __name__ == '__main__':
    unittest.main()



import csv
import logging
import os
from abc import ABC
from multiprocessing import Process
from multiprocessing.queues import Queue
from queue import Empty
from typing import List

from galileodb.db import ExperimentDatabase
from galileodb.model import RequestTrace
from galileodb.reporter.traces import RedisTraceReporter
from galileodb.sql.adapter import ExperimentSQLDatabase

logger = logging.getLogger(__name__)

POISON = "__POISON__"
START = "__START__"
PAUSE = "__PAUSE__"
FLUSH = '__FLUSH__'


class TraceWriter(ABC):
    def write(self, traces: List[RequestTrace]):
        raise NotImplementedError


class TraceLogger(Process):
    flush_interval = 20

    def __init__(self, trace_queue: Queue, writer: TraceWriter = None, start=True) -> None:
        super().__init__()
        self.traces = trace_queue
        self.closed = False
        self.buffer = list()
        self.writer = writer
        self.running = start

    def run(self):
        try:
            return self.listen()
        finally:
            self.flush()

    def flush(self):
        if not self.buffer:
            logger.debug('buffer empty, not flushing')
            return

        logger.debug('flushing trace buffer')

        if self.writer:
            try:
                self.writer.write(self.buffer)
            except Exception as e:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception('error writing traces')
                else:
                    logger.error('error writing traces: %s', e)
        else:
            logger.debug('no writer to flush to')

        self.buffer.clear()

    def close(self):
        self.closed = True
        self.traces.put(POISON)

    def listen(self):
        timeout = None
        while True:
            if self.closed and timeout is None:
                logger.debug('setting read timeout to 2 seconds')
                timeout = 2

            try:
                trace = self.traces.get(timeout=timeout)

                if trace == POISON:
                    logger.debug('poison received, setting closed to true')
                    self.closed = True
                    break
                elif trace == FLUSH:
                    logger.debug('flush command received, flushing buffer')
                    self.flush()
                    continue
                elif trace == START:
                    logger.debug('start received')
                    self.running = True
                    continue
                elif trace == PAUSE:
                    logger.debug('pause received, flushing remaining traces')
                    self.running = False
                    self.flush()
                    continue

                if self.running:
                    self.buffer.append(trace)

                if len(self.buffer) >= self.flush_interval:
                    logger.debug('flush interval reached, flushing buffer')
                    self.flush()

            except KeyboardInterrupt:
                break
            except Empty:
                logger.debug('queue is empty, exiting')
                return


class RedisTopicTraceWriter(TraceWriter):

    def __init__(self, rds) -> None:
        super().__init__()
        self.reporter = RedisTraceReporter(rds)

    def write(self, traces: List[RequestTrace]):
        self.reporter.report_multiple(traces)


class DatabaseTraceWriter(TraceWriter):
    experiment_db: ExperimentDatabase

    def __init__(self, experiment_db: ExperimentDatabase) -> None:
        self.experiment_db = experiment_db
        self.connected = False

    def write(self, traces: List[RequestTrace]):
        self._assert_connection()
        self.experiment_db.save_traces(traces)

    def _assert_connection(self):
        # this is a terrible hack due to multiprocessing issues:
        # close() will delete the threadlocal (which is not actually accessible from the process) and create a new
        # connection. The SqlAdapter adapter design may be broken. or python multiprocessing...
        if self.connected:
            return

        if isinstance(self.experiment_db, ExperimentSQLDatabase):
            self.experiment_db.db.reconnect()
            self.connected = True


class FileTraceWriter(TraceWriter):

    def __init__(self, host_name, target_dir='/tmp/mc2/exp') -> None:
        self.target_dir = target_dir
        self.file_name = 'traces-%s.csv' % host_name
        self.file_path = os.path.join(self.target_dir, self.file_name)
        self.mkdirp(self.target_dir)
        self.init_file()

    def init_file(self):
        logger.debug('initializing trace file logger to log into %s', self.file_path)
        if os.path.exists(self.file_path):
            return

        logger.debug('initializing %s with header', self.file_path)
        with open(self.file_path, 'w') as fd:
            csv.writer(fd).writerow(RequestTrace._fields)

    def write(self, buffer: List[RequestTrace]):
        with open(self.file_path, 'a') as fd:
            writer = csv.writer(fd)
            for row in buffer:
                writer.writerow(row)

    @staticmethod
    def mkdirp(path):
        if not os.path.exists(path):
            os.makedirs(path)

        if os.path.isfile(path):
            raise FileExistsError("%s is an existing file" % path)



import os
import shutil
import tempfile
import time
from queue import Queue
from threading import Thread, Event

import redislite

from galileodb.sql.driver.sqlite import SqliteAdapter
from tests.sql.adapter import AbstractTestSqlDatabase


def poll(condition, timeout=None, interval=0.5):
    remaining = 0
    if timeout is not None:
        remaining = timeout

    while not condition():
        if timeout is not None:
            remaining -= interval

            if remaining <= 0:
                raise TimeoutError('gave up waiting after %s seconds' % timeout)

        time.sleep(interval)


def assert_poll(condition, msg='Condition failed'):
    try:
        poll(condition, 2, 0.01)
    except TimeoutError:
        raise AssertionError(msg)


class TestResource(object):

    def setUp(self):
        pass

    def tearDown(self):
        pass


class RedisResource(TestResource):
    tmpfile: str
    rds: redislite.Redis

    def setUp(self):
        self.tmpfile = tempfile.mktemp('.db', 'galileo_test_')
        self.rds = redislite.Redis(self.tmpfile, decode_responses=True)
        self.rds.get('dummykey')  # run a first command to initiate

    def tearDown(self):
        self.rds.shutdown()

        os.remove(self.tmpfile)
        os.remove(self.rds.redis_configuration_filename)
        os.remove(self.rds.settingregistryfile)
        shutil.rmtree(self.rds.redis_dir)

        self.rds = None
        self.tmpfile = None


class SqliteResource(AbstractTestSqlDatabase, TestResource):
    db_file = None

    def setUp(self) -> None:
        self.db_file = tempfile.mktemp('.sqlite', 'galileo_test_')
        super().setUp()

    def _create_sql_adapter(self):
        return SqliteAdapter(self.db_file)

    def tearDown(self) -> None:
        super().tearDown()
        os.remove(self.db_file)


class RedisSubscriber:

    def __init__(self, rds, channel, queue=None) -> None:
        super().__init__()
        self.rds = rds
        self.channel = channel
        self.queue = queue or Queue()
        self.pubsub = rds.pubsub()
        self.t = None

        self._listening = Event()

    def start(self) -> Thread:
        if not self.t:
            self.t = Thread(target=self.listen)
            self.t.start()
            self._listening.wait(2)

        return self.t

    def join(self, timeout):
        if self.t:
            self.t.join(timeout)

    def listen(self):
        self.pubsub.subscribe(self.channel)
        try:
            for item in self.pubsub.listen():
                if item['type'] == 'subscribe':
                    self._listening.set()
                elif item['type'] == 'message':
                    self.queue.put(item['data'])
                else:
                    print('ignoring pubsub data', item)
        except:
            return

    def close(self):
        self.pubsub.close()

    def shutdown(self, timeout=1):
        self.close()
        self.join(timeout)



from galileodb.model import Experiment, Telemetry
from galileodb.sql.adapter import ExperimentSQLDatabase, SqlAdapter
from tests.test_db import AbstractTestExperimentDatabase


class AbstractTestSqlDatabase(AbstractTestExperimentDatabase):
    sql: SqlAdapter
    db: ExperimentSQLDatabase

    def setUp(self) -> None:
        self.sql = self._create_sql_adapter()
        self.db = ExperimentSQLDatabase(self.sql)
        self.db.open()

    def tearDown(self) -> None:
        self.db.close()

    def _create_sql_adapter(self) -> SqlAdapter:
        raise NotImplementedError

    def test_insert_many_and_count(self):
        entries = [
            ('expid3', 'test_experiment', 'unittest2', 'running'),
            ('expid4', 'test_experiment', 'unittest2', 'running'),
            ('expid5', 'test_experiment', 'unittest2', 'running')
        ]

        self.sql.insert_many('experiments', ['exp_id', 'name', 'creator', 'status'], entries)

        cur = self.sql.cursor()
        cur.execute('SELECT COUNT(*) FROM experiments WHERE `CREATOR` = "unittest2"')
        val = cur.fetchone()
        self.assertEqual(3, val[0])

    def test_update_and_get(self):
        self.db.save_experiment(Experiment('expid6', 'test_experiment', 'unittest', 10, 100, 1, 'finished'))
        self.db.save_experiment(Experiment('expid7', 'test_experiment', 'unittest', 100, None, 10, 'running'))

        self.sql.update_by_id('experiments', ('exp_id', 'expid7'), {'status': 'finished', 'end': 1000})

        exp2 = self.db.get_experiment('expid7')

        self.assertEqual('expid7', exp2.id)
        self.assertEqual('test_experiment', exp2.name)
        self.assertEqual('unittest', exp2.creator)
        self.assertEqual(100., exp2.start)
        self.assertEqual(1000., exp2.end)
        self.assertEqual(10., exp2.created)
        self.assertEqual('finished', exp2.status)

    def test_save_telemetry_writes_to_table(self):
        telemetry = [
            Telemetry(1, 'cpu', 'n1', 32, 'expid1'),
            Telemetry(2, 'cpu', 'n1', 33, 'expid1'),
            Telemetry(3, 'cpu', 'n1', 31, 'expid1'),
        ]

        self.db.save_telemetry(telemetry)

        rows = self.db.db.fetchall('SELECT * FROM telemetry')
        self.assertEqual(3, len(rows))

    def test_delete_experiment_removes_telemetry(self):
        exp_id = 'expid10'
        exp_id_control = 'expid11'
        self.db.save_experiment(Experiment(exp_id, 'test_experiment', 'unittest', 10, 100, 1, 'finished'))
        self.db.save_telemetry([Telemetry(1, 'cpu', 'n1', 32, exp_id)])

        self.db.save_experiment(Experiment(exp_id_control, 'test_experiment', 'unittest', 10, 100, 1, 'finished'))
        self.db.save_telemetry([Telemetry(1, 'cpu', 'n1', 32, exp_id_control)])

        self.assertIsNotNone(self.db.get_experiment(exp_id))
        telemetry_rows = self.db.db.fetchall('SELECT * FROM telemetry WHERE EXP_ID = "%s"' % exp_id)
        self.assertEqual(1, len(telemetry_rows))

        self.db.delete_experiment(exp_id)

        self.assertIsNone(self.db.get_experiment(exp_id))
        telemetry_rows = self.db.db.fetchall('SELECT * FROM telemetry WHERE EXP_ID = "%s"' % exp_id)
        self.assertEqual(0, len(telemetry_rows))

        self.assertIsNotNone(self.db.get_experiment(exp_id_control))
        telemetry_rows = self.db.db.fetchall('SELECT * FROM telemetry WHERE EXP_ID = "%s"' % exp_id_control)
        self.assertEqual(1, len(telemetry_rows))



from galileodb.db import ExperimentDatabase
from galileodb.model import Experiment, Telemetry, NodeInfo

name = 'galileodb'

__all__ = [
    'ExperimentDatabase',
    'Experiment',
    'Telemetry',
    'NodeInfo'
]



import logging

import mysql.connector as mysql

from galileodb.sql.adapter import SqlAdapter

logger = logging.getLogger(__name__)


class MysqlAdapter(SqlAdapter):
    placeholder = '%s'

    def _connect(self, *args, **kwargs):
        logger.info('connecting to mysql db with args %s', kwargs)
        con = mysql.connect(**kwargs)
        con.autocommit = True
        return con

    def cursor(self):
        try:
            return self.db.cursor()
        except:
            logger.warning('tyring to reconnect...')
            self.reconnect()
            return self.db.cursor()

    def executescript(self, *args, **kwargs):
        script = args[0]
        # FIXME: pretty hacky
        schema = ' '.join(script.splitlines())
        statements = [stmt.strip() for stmt in schema.split(';')]

        cur = self.db.cursor()
        try:
            for stmt in statements:
                if stmt:
                    cur.execute(stmt)
            self.db.commit()
        except:
            self.db.rollback()



import multiprocessing
import shutil
import threading
import unittest
from time import sleep
from unittest.mock import patch

from timeout_decorator import timeout_decorator

from galileodb.model import RequestTrace
from galileodb.reporter.traces import RedisTraceReporter
from galileodb.trace import TraceLogger, POISON, START, PAUSE, FLUSH, TraceWriter, FileTraceWriter, \
    RedisTopicTraceWriter, DatabaseTraceWriter
from tests.testutils import RedisResource, SqliteResource, assert_poll, RedisSubscriber

traces = [
    RequestTrace('req1', 'client', 'service', 1.1, 1.2, 1.3),
    RequestTrace('req2', 'client', 'service', 2.2, 2.3, 2.4, 200, 'server1', 'exp1','{"headers": 1}', 'hello'),
    RequestTrace('req3', 'client', 'service', 3.2, 3.3, 3.4, status=200, server='server1',
                 response='foo=bar\nx=1,"2",3')
]


class TestTraceLogger(unittest.TestCase):

    def setUp(self) -> None:
        self.queue = multiprocessing.Queue()
        self.logger = TraceLogger(self.queue)
        self.logger.flush_interval = 2
        self.flush_interval = 2
        self.thread = threading.Thread(target=self.logger.run)
        self.thread.start()

    def tearDown(self) -> None:
        if not self.logger.closed:
            self.logger.close()
        self.thread.join(2)

    @patch('galileodb.trace.TraceLogger.flush')
    def test_flush_called(self, flush):
        self.trigger_flush()
        assert_poll(lambda: flush.called_once(), 'Flush was not called after triggering flush')

    @patch('galileodb.trace.TraceLogger.flush')
    def test_flush_called_after_flush_msg(self, flush):
        self.send_message(FLUSH)
        assert_poll(lambda: flush.called_once(), 'Flush was not called after triggering FLUSH')

    def test_closed_after_poison(self):
        self.send_message(POISON)
        assert_poll(lambda: self.logger.closed, 'POISON didnt close logger')

    @patch('galileodb.trace.TraceLogger.flush')
    def test_flushed_after_poison(self, flush):
        self.send_message(POISON)
        assert_poll(lambda: flush.called_once(), 'Flush was not called after POISON')

    def test_running_after_start(self):
        self.send_message(START)
        assert_poll(lambda: self.logger.running, 'Logger not running after START')

    def test_not_running_after_pause(self):
        self.send_message(PAUSE)
        assert_poll(lambda: not self.logger.running, 'Logger running after PAUSE')

    @patch('galileodb.trace.TraceLogger.flush')
    def test_discarding_messages_in_paused_state(self, flush):
        self.send_message(PAUSE)
        assert_poll(lambda: not self.logger.running, 'Logger running after PAUSE')

        self.queue.put(traces[0])
        self.queue.put(traces[1])

        sleep(0.5)
        self.assertEqual(0, len(self.logger.buffer), 'buffer should not be filled when logger is paused')

        self.send_message(START)
        assert_poll(lambda: self.logger.running, 'Logger not running after START')

        self.queue.put(traces[2])
        assert_poll(lambda: len(self.logger.buffer) == 1, 'buffer should contain one message after started')

    def test_flush_calls_writer(self):
        written = list()

        class DummyWriter(TraceWriter):
            def write(self, buffer):
                written.extend(buffer)

        self.logger.writer = DummyWriter()

        self.trigger_flush()

        assert_poll(lambda: len(written) == self.flush_interval)

    def trigger_flush(self):
        for i in range(self.flush_interval):
            self.queue.put(RequestTrace(f'req{i}', 'client', 'service', 1, 1, 1, status=200, response='data'))

    def send_message(self, msg):
        self.queue.put(msg)


class TestDatabaseTraceWriter(unittest.TestCase):
    sql_resource = SqliteResource()

    def setUp(self) -> None:
        self.sql_resource.setUp()

    def tearDown(self) -> None:
        self.sql_resource.tearDown()

    @timeout_decorator.timeout(5)
    def test_write(self):
        self.writer = DatabaseTraceWriter(self.sql_resource.db)
        self.writer.write(traces)

        actual = self.sql_resource.db.get_traces()

        self.assertEqual(3, len(actual))

        self.assertEqual(traces[0], actual[0])
        self.assertEqual(traces[1], actual[1])
        self.assertEqual(traces[2], actual[2])


class TestRedisTopicTraceWriter(unittest.TestCase):
    redis_resource = RedisResource()
    writer: RedisTopicTraceWriter

    def setUp(self) -> None:
        self.redis_resource.setUp()

        self.writer = RedisTopicTraceWriter(self.redis_resource.rds)
        self.sub = RedisSubscriber(self.redis_resource.rds, RedisTraceReporter.channel)
        self.sub.start()

    def tearDown(self) -> None:
        self.sub.shutdown()
        self.redis_resource.tearDown()

    @timeout_decorator.timeout(5)
    def test_write(self):
        self.writer.write(traces)

        t1 = self.sub.queue.get(timeout=2)
        self.assertEqual(t1, 'req1,client,service,1.1000000,1.2000000,1.3000000,-1,None,None,None,None')
        t2 = self.sub.queue.get(timeout=2)
        self.assertEqual(t2, 'req2,client,service,2.2000000,2.3000000,2.4000000,200,server1,exp1,{"headers": 1},hello')
        t3 = self.sub.queue.get(timeout=2)
        self.assertEqual(t3, r'req3,client,service,3.2000000,3.3000000,3.4000000,200,server1,None,None,foo=bar\nx=1,"2",3')


class TestFileTraceWriter(unittest.TestCase):
    target_dir = '/tmp/galileo_test'

    def setUp(self) -> None:
        self.writer = FileTraceWriter('test', self.target_dir)

    def tearDown(self) -> None:
        shutil.rmtree(self.target_dir)

    def test_write(self):
        self.writer.write(traces[:1])
        self.writer.write(traces[1:])  # makes sure consecutive write calls append to file

        expected = \
            'request_id,client,service,created,sent,done,status,server,exp_id,headers,response\n' + \
            'req1,client,service,1.1,1.2,1.3,-1,,,,\n' + \
            'req2,client,service,2.2,2.3,2.4,200,server1,exp1,"{""headers"": 1}",hello\n' + \
            'req3,client,service,3.2,3.3,3.4,200,server1,,,"foo=bar\n' + \
            'x=1,""2"",3"'  # this is just how the csv writer resolves the double quotes ...

        with open(self.writer.file_path) as fd:
            actual = fd.read()

        self.assertEqual(expected.strip(), actual.strip())



import logging
import os
from typing import MutableMapping

from galileodb.db import ExperimentDatabase
from galileodb.mixed.db import MixedExperimentDatabase
from galileodb.sql.adapter import ExperimentSQLDatabase

logger = logging.getLogger(__name__)


def create_experiment_database_from_env(env: MutableMapping = os.environ) -> ExperimentDatabase:
    driver = env.get('galileo_expdb_driver', 'sqlite')
    return create_experiment_database(driver)


def create_experiment_database(driver: str, env: MutableMapping = os.environ) -> ExperimentDatabase:
    if driver == 'influxdb':
        return create_influxdb_from_env(env)

    from galileodb.sql.adapter import ExperimentSQLDatabase

    if driver == 'mixed':
        return create_mixeddb_from_env(env)

    if driver == 'sqlite':
        db_adapter = create_sqlite_from_env(env)

    elif driver == 'mysql':
        db_adapter = create_mysql_from_env(env)

    else:
        raise ValueError('unknown database driver %s' % driver)

    return ExperimentSQLDatabase(db_adapter)


def create_influxdb_from_env(env: MutableMapping = os.environ):
    from galileodb.influx.db import InfluxExperimentDatabase
    from influxdb_client import InfluxDBClient

    params = {
        'url': env.get('galileo_expdb_influxdb_url', 'http://localhost:8086'),
        'token': env.get('galileo_expdb_influxdb_token', 'my-token'),
        'timeout': int(env.get('galileo_expdb_influxdb_timeout', '10000')),
        'org': env.get('galileo_expdb_influxdb_org', 'galileo'),
        'org_id': env.get('galileo_expdb_influxdb_org_id', 'org-id')
    }

    return InfluxExperimentDatabase(InfluxDBClient(**params), org_name=params['org'], org_id=params['org_id'])


def create_mysql_from_env(env: MutableMapping = os.environ):
    from galileodb.sql.driver.mysql import MysqlAdapter
    params = {
        'host': env.get('galileo_expdb_mysql_host', 'localhost'),
        'port': int(env.get('galileo_expdb_mysql_port', '3307')),
        'user': env.get('galileo_expdb_mysql_user', None),
        'password': env.get('galileo_expdb_mysql_password', None),
        'db': env.get('galileo_expdb_mysql_db', None)
    }

    logger.info('read mysql adapter parameters from environment %s', params)
    return MysqlAdapter(**params)


def create_sqlite_from_env(env: MutableMapping = os.environ):
    from galileodb.sql.driver.sqlite import SqliteAdapter
    db_file = env.get('galileo_expdb_sqlite_path', './galileodb.sqlite')

    logger.info('creating db adapter to SQLite %s', os.path.realpath(db_file))
    return SqliteAdapter(db_file)


def create_mixeddb_from_env(env: MutableMapping = os.environ):
    influxdb = create_influxdb_from_env(env)
    mysql_adapter = create_mysql_from_env(env)
    sqldb = ExperimentSQLDatabase(mysql_adapter)

    return MixedExperimentDatabase(influxdb, sqldb)



import unittest
from queue import Queue
from typing import List

from galileodb.model import RequestTrace
from galileodb.recorder.traces import TraceRecorder, TracesSubscriber, RedisTraceRecorder
from galileodb.reporter.traces import RedisTraceReporter
from galileodb.trace import TraceWriter
from tests.testutils import RedisResource


class TraceRecorderTest(unittest.TestCase):
    redis = RedisResource()

    def setUp(self) -> None:
        self.redis.setUp()

    def tearDown(self) -> None:
        self.redis.tearDown()

    def test_recorder(self):
        queue = Queue()

        class TestTraceRecorder(TraceRecorder):
            def _record(self, t):
                queue.put(t)

        recorder = TestTraceRecorder(self.redis.rds)
        recorder.start()

        headers = '{"server": ["Server", "BaseHTTP/0.6 Python/3.9.5"], "date": ["Date", "Tue, 14 Sep 2021 08:54:40 GMT"], "content-type": ["Content-type", "text/html"]}'
        reporter = RedisTraceReporter(self.redis.rds)
        fixture = [
            RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, response='hello there', headers=headers),
            RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, server='s1', exp_id='exp1')
        ]
        reporter.report_multiple(fixture)

        trace = queue.get(timeout=2)
        self.assertEqual(fixture[0], trace)

        trace = queue.get(timeout=2)
        self.assertEqual(fixture[1], trace)

        recorder.stop(timeout=2)


class TraceSubscriberTest(unittest.TestCase):
    redis = RedisResource()

    def setUp(self) -> None:
        self.redis.setUp()

    def tearDown(self) -> None:
        self.redis.tearDown()

    def test_parse(self):
        parse = TracesSubscriber.parse

        expected = RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, response="foo=bar\nfield1=value1,a,b,c")
        actual = parse(r"r1,c1,s1,1.1000000,1.2000000,1.3000000,200,None,None,None,foo=bar\nfield1=value1,a,b,c")

        self.assertEqual(expected, actual)

    def test_parse_with_headers(self):
        parse = TracesSubscriber.parse
        headers = '{"server": ["Server", "BaseHTTP/0.6 Python/3.9.5"], "date": ["Date", "Tue, 14 Sep 2021 08:54:40 GMT"], "content-type": ["Content-type", "text/html"]}'
        expected = RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, response='foo=bar\nfield1=value1,a,b,c',
                                headers=headers)
        actual = parse(
            fr"r1,c1,s1,1.1000000,1.2000000,1.3000000,200,None,None,{headers.replace(',', '|')},foo=bar\nfield1=value1,a,b,c")
        self.assertEqual(expected, actual)


class RedisTraceRecorderTest(unittest.TestCase):
    redis = RedisResource()

    def setUp(self) -> None:
        self.redis.setUp()

    def tearDown(self) -> None:
        self.redis.tearDown()

    # @unittest.skip("FIXME: sometimes fails")
    def test_recorder(self):
        queue = Queue()

        class TestTraceWriter(TraceWriter):
            def write(self, traces: List[RequestTrace]):
                for t in traces:
                    queue.put(t)

        recorder = RedisTraceRecorder(self.redis.rds, exp_id='exp1', writer=TestTraceWriter(), flush_every=2)
        recorder.start()

        reporter = RedisTraceReporter(self.redis.rds)
        headers = '\'\"{"server": ["Server", "BaseHTTP/0.6 Python/3.9.5"], "date": ["Date", "Tue, 14 Sep 2021 08:54:40 GMT"], "content-type": ["Content-type", "text/html"]}\"\''
        fixture = [
            RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, headers=headers, response='hello there'),
            RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, server='s1'),
        ]

        expected = [
            RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, headers=headers, response='hello there', exp_id='exp1'),
            RequestTrace('r1', 'c1', 's1', 1.1, 1.2, 1.3, 200, server='s1', exp_id='exp1')
        ]

        reporter.report_multiple(fixture)

        trace = queue.get(timeout=2)
        self.assertEqual(expected[0], trace)

        trace = queue.get(timeout=2)
        self.assertEqual(expected[1], trace)

        recorder.stop(timeout=2)



import os

import setuptools

try:
    with open("README.md", "r") as fh:
        long_description = fh.read()
except FileNotFoundError:
    long_description = ''

try:
    with open("requirements-dev.txt", "r") as fh:
        tests_require = [line for line in fh.read().split(os.linesep) if line]
except FileNotFoundError:
    tests_require = []

try:
    with open("requirements.txt", "r") as fh:
        install_requires = [line for line in fh.read().split(os.linesep) if line and not line.startswith('git')]
except FileNotFoundError:
    install_requires = []

setuptools.setup(
    name="galileo-db",
    version="0.10.2",
    author="Thomas Rausch",
    author_email="t.rausch@dsg.tuwien.ac.at",
    description="Galileo DB: Gateway and client tools for the Galileo Experiment DB",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/edgerun/galileo-db",
    packages=setuptools.find_packages(),
    package_data={'galileodb.sql': ['schema.sql']},
    test_suite="tests",
    tests_require=tests_require,
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            'galileodb-recorder = galileodb.cli.recorder:main',
            'galileodb-ctl = galileodb.cli.manager:main'
        ]
    },
)



import abc

from galileodb import ExperimentDatabase, Experiment, Telemetry
from galileodb.model import ExperimentEvent, RequestTrace


class AbstractTestExperimentDatabase(abc.ABC):
    db: ExperimentDatabase

    def test_get_experiment_on_invalid_id(self):
        result = self.db.get_experiment('NOTEXISTS')
        self.assertIsNone(result)

    def test_save_and_get_experiment(self):
        expected = Experiment('expid1', 'test_experiment', 'unittest', 10, 100, 1, 'running')

        self.db.save_experiment(expected)

        actual = self.db.get_experiment('expid1')

        self.assertEqual('expid1', actual.id)
        self.assertEqual('test_experiment', actual.name)
        self.assertEqual('unittest', actual.creator)
        self.assertEqual(10., actual.start)
        self.assertEqual(100., actual.end)
        self.assertEqual(1., actual.created)
        self.assertEqual('running', actual.status)

    def test_save_experiment_and_get(self):
        entity = Experiment('exp1', 'exp1-name', 'exp1-creator', 1.1, 2.2, status='FINISHED')
        self.db.save_experiment(entity)
        actual = self.db.get_experiment('exp1')
        self.assertEqual(entity.__dict__, actual.__dict__)

    def test_save_experiment_and_find_all(self):
        entity1 = Experiment('exp1', 'exp1-name', 'exp1-creator', 1.1, 2.2, status='FINISHED')
        entity2 = Experiment('exp2', 'exp2-name', 'exp2-creator', 3.3, 4.4, status='RUNNING')
        self.db.save_experiment(entity1)
        self.db.save_experiment(entity2)

        experiments = self.db.find_all()
        self.assertEqual(2, len(experiments))
        self.assertEqual(entity1.__dict__, experiments[0].__dict__)
        self.assertEqual(entity2.__dict__, experiments[1].__dict__)

    def test_get_running_experiment(self):
        entity1 = Experiment('exp1', 'exp1-name', 'exp1-creator', 1.1, 2.2, status='FINISHED')
        entity2 = Experiment('exp2', 'exp2-name', 'exp2-creator', 3.3, 4.4, status='RUNNING')
        self.db.save_experiment(entity1)
        self.db.save_experiment(entity2)

        running = self.db.get_running_experiment()

        self.assertEqual('exp2', running.id)

    def test_get_running_experiment_on_empty_set(self):
        entity1 = Experiment('exp1', 'exp1-name', 'exp1-creator', 1.1, 2.2, status='FINISHED')
        entity2 = Experiment('exp2', 'exp2-name', 'exp2-creator', 3.3, 4.4, status='FINISHED')
        self.db.save_experiment(entity1)
        self.db.save_experiment(entity2)

        running = self.db.get_running_experiment()

        self.assertIsNone(running)

    def test_update_experiment_and_get(self):
        exp = Experiment('expid8', 'test_experiment', 'unittest', 10, None, 1, 'running')
        self.db.save_experiment(exp)

        self.assertEqual(self.db.get_experiment('expid8').status, 'running')
        self.assertEqual(self.db.get_experiment('expid8').end, None)

        exp.status = 'finished'
        exp.end = 100

        self.db.update_experiment(exp)

        self.assertEqual(self.db.get_experiment('expid8').status, 'finished')
        self.assertEqual(self.db.get_experiment('expid8').end, 100)

    def test_save_and_get_telemetry(self):
        telemetry = [
            Telemetry(1, 'cpu', 'n1', 32, 'expid1'),
            Telemetry(2, 'cpu', 'n1', 33, 'expid1'),
            Telemetry(3, 'cpu', 'n1', 31, 'expid2'),
            Telemetry(4, 'rx', 'n1', 32, 'expid2', 'eth0'),
        ]

        self.db.save_telemetry(telemetry)

        actual = self.db.get_telemetry('expid1')
        self.assertEqual(2, len(actual))
        self.assertEqual(telemetry[0], actual[0])
        self.assertEqual(telemetry[1], actual[1])

        actual = self.db.get_telemetry('expid2')
        self.assertEqual(2, len(actual))
        self.assertEqual(telemetry[2], actual[0])
        self.assertEqual(telemetry[3], actual[1])

        actual = self.db.get_telemetry()
        self.assertEqual(4, len(actual))
        self.assertEqual(telemetry[0], actual[0])
        self.assertEqual(telemetry[1], actual[1])
        self.assertEqual(telemetry[2], actual[2])
        self.assertEqual(telemetry[3], actual[3])

    def test_save_and_touch_and_get_traces(self):
        traces = [
            RequestTrace('req1', 'c1', 's1', 1.1, 1.2, 1.3, server='h1', status=200),
            RequestTrace('req2', 'c2', 's2', 2.1, 2.2, 2.3, server='h2', status=200),
            RequestTrace('req3', 'c3', 's1', 3.1, 3.2, 3.3, server='h1', status=200, response='time=123'),
            RequestTrace('req4', 'c1', 's1', 4.1, 4.2, 4.3, server='h1', status=200),
        ]

        self.db.save_experiment(Experiment('exp1', start=1, end=3.5, status='FINISHED'))

        self.db.save_traces(traces)

        self.db.touch_traces(self.db.get_experiment('exp1'))  # should touch the first 3

        actual = self.db.get_traces('exp1')

        expected = [
            RequestTrace('req1', 'c1', 's1', 1.1, 1.2, 1.3, server='h1', status=200, exp_id='exp1'),
            RequestTrace('req2', 'c2', 's2', 2.1, 2.2, 2.3, server='h2', status=200, exp_id='exp1'),
            RequestTrace('req3', 'c3', 's1', 3.1, 3.2, 3.3, server='h1', status=200, response='time=123', exp_id='exp1')
        ]

        self.assertEqual(expected, actual)

    def test_save_and_get_events(self):
        events = [
            ExperimentEvent('exp1', 1, 'begin'),
            ExperimentEvent('exp1', 2, 'start', 'function1'),
            ExperimentEvent('exp2', 3, 'stop', 'function1'),
        ]

        self.db.save_events(events)

        stored = self.db.get_events()
        self.assertEqual(3, len(stored))

        self.assertEqual(ExperimentEvent('exp1', 1.0, 'begin', None), stored[0])
        self.assertEqual(ExperimentEvent('exp1', 2.0, 'start', 'function1'), stored[1])
        self.assertEqual(ExperimentEvent('exp2', 3.0, 'stop', 'function1'), stored[2])

    def test_save_and_get_events_for_experiment(self):
        events = [
            ExperimentEvent('exp1', 1, 'begin'),
            ExperimentEvent('exp1', 2, 'start', 'function1'),
            ExperimentEvent('exp2', 3, 'stop', 'function1'),
        ]

        self.db.save_events(events)

        stored = self.db.get_events('exp1')
        self.assertEqual(2, len(stored))

        self.assertEqual(ExperimentEvent('exp1', 1.0, 'begin', None), stored[0])
        self.assertEqual(ExperimentEvent('exp1', 2.0, 'start', 'function1'), stored[1])

    def test_delete_experiment(self):
        exp_id = 'expid10'
        exp_id_control = 'expid11'
        self.db.save_experiment(Experiment(exp_id, 'test_experiment', 'unittest', 10, 100, 1, 'finished'))
        self.db.save_telemetry([Telemetry(1, 'cpu', 'n1', 32, exp_id)])

        self.db.save_experiment(Experiment(exp_id_control, 'test_experiment', 'unittest', 10, 100, 1, 'finished'))
        self.db.save_telemetry([Telemetry(1, 'cpu', 'n1', 32, exp_id_control)])

        self.assertIsNotNone(self.db.get_experiment(exp_id))

        self.db.delete_experiment(exp_id)

        self.assertIsNone(self.db.get_experiment(exp_id))
        self.assertIsNotNone(self.db.get_experiment(exp_id_control))
